@@ -14,15 +14,15 @@ Author: Jonathan Slenders
 """
 from __future__ import unicode_literals
 
-import codecs
 import fcntl
 import os
 import select
-import six
 import sys
 import errno
 import threading
 
+from codecs import getincrementaldecoder
+
 from .code import Code
 from .inputstream import InputStream
 from .inputstream_handler import InputStreamHandler
@@ -91,6 +91,8 @@ class CommandLine(object):
     #: When to call the `on_input_timeout` callback.
     input_timeout = .5
 
+    stdin_decoder_cls = getincrementaldecoder('utf-8')
+
     def __init__(self, stdin=None, stdout=None):
         self.stdin = stdin or sys.stdin
         self.stdout = stdout or sys.stdout
@@ -112,6 +114,11 @@ class CommandLine(object):
         #: Currently reading input.
         self.is_reading_input = False
 
+        # Create incremental decoder for decoding stdin.
+        # We can not just do `os.read(stdin.fileno(), 1024).decode('utf-8')`, because
+        # it could be that we are in the middle of a utf-8 byte sequence.
+        self._stdin_decoder = self.stdin_decoder_cls()
+
     def request_redraw(self):
         """
         Thread safe way of sending a repaint trigger to the input event loop.
@@ -190,7 +197,21 @@ class CommandLine(object):
         #       `codecs.getreader('utf-8')(stdin)` and doing `read(1)`.
         #       Somehow that causes some latency when the escape
         #       character is pressed. (Especially on combination with the `select`.
-        return os.read(self.stdin.fileno(), 1024).decode('utf-8')
+        try:
+            bytes = os.read(self.stdin.fileno(), 1024)
+        except OSError:
+            # In case of SIGWINCH
+            bytes = b''
+
+        try:
+            return self._stdin_decoder.decode(bytes)
+        except UnicodeDecodeError:
+            # When it's not possible to decode this bytes, reset the decoder.
+            # The only occurence of this that I had was when using iTerm2 on OS
+            # X, with "Option as Meta" checked (You should choose "Option as
+            # +Esc".)
+            self._stdin_decoder = self.stdin_decoder_cls()
+            return ''
 
     def read_input(self, initial_value='', on_abort=AbortAction.RETRY, on_exit=AbortAction.IGNORE):
         """
@@ -305,7 +326,7 @@ def _select(*args, **kwargs):
     while True:
         try:
             return select.select(*args, **kwargs)
-        except Exception as e:
+        except select.error as e:
             # Retry select call when EINTR
             if e.args and e.args[0] == errno.EINTR:
                 continue
