@@ -109,6 +109,9 @@ class CompletionState(object):
         #: Document as it was when the completion started.
         self.original_document = original_document
 
+        self.original_text_before_cursor = original_document.text_before_cursor
+        self.original_text_after_cursor = original_document.text_after_cursor
+
         #: List of all the current Completion instances which are possible at
         #: this point.
         self.current_completions = current_completions or []
@@ -121,12 +124,19 @@ class CompletionState(object):
     def original_cursor_position(self):
         self.original_document.cursor_position
 
-    @property
-    def current_completion_text(self):
+    def get_new_text_and_position(self):
+        """ Return (new_text, new_cursor_position) for this completion. """
         if self.complete_index is None:
-            return ''
+            return self.original_document.text, self.original_document.cursor_position
         else:
-            return self.current_completions[self.complete_index].suffix
+            c = self.current_completions[self.complete_index]
+            if c.start_position == 0:
+                before = self.original_text_before_cursor
+            else:
+                before = self.original_text_before_cursor[:c.start_position]
+            return before + c.text + self.original_text_after_cursor, len(before) + len(c.text)
+
+
 
 class _IncrementalSearchState(object):
     def __init__(self, original_cursor_position, original_working_index):
@@ -261,12 +271,12 @@ class Line(object):
         Safe current state (input text and cursor position), so that we can
         restore it by calling undo.
         """
-        state = (self.text, self.cursor_position)
-
         # Safe if the text is different from the text at the top of the stack
-        # is different.
-        if not self._undo_stack or self._undo_stack[-1][0] != state[0]:
-            self._undo_stack.append(state)
+        # is different. If the text is the same, just update the cursor position.
+        if self._undo_stack and self._undo_stack[-1][0] == self.text:
+            self._undo_stack[-1] = (self._undo_stack[-1][0], self.cursor_position)
+        else:
+            self._undo_stack.append((self.text, self.cursor_position))
 
     def set_current_line(self, value):
         """
@@ -323,41 +333,23 @@ class Line(object):
         """
         self._arg_prompt_text = arg
 
-    @_to_mode(LineMode.NORMAL)
-    def home(self):
-        self.cursor_position = 0
-
-    @_to_mode(LineMode.NORMAL)
-    def end(self):
-        self.cursor_position = len(self.text)
-
     @_to_mode(LineMode.NORMAL)
     def cursor_left(self):
-        if self.document.cursor_position_col > 0:
-            self.cursor_position -= 1
+        self.cursor_position += self.document.get_cursor_left_position()
 
     @_to_mode(LineMode.NORMAL)
     def cursor_right(self):
-        if not self.document.cursor_at_the_end_of_line:
-            self.cursor_position += 1
+        self.cursor_position += self.document.get_cursor_right_position()
 
     @_to_mode(LineMode.NORMAL)
     def cursor_up(self):
-        """
-        (for multiline edit). Move cursor to the previous line.
-        """
-        new_pos = self.document.cursor_up_position
-        if new_pos is not None:
-            self.cursor_position = new_pos
+        """ (for multiline edit). Move cursor to the previous line.  """
+        self.cursor_position += self.document.get_cursor_up_position()
 
     @_to_mode(LineMode.NORMAL)
     def cursor_down(self):
-        """
-        (for multiline edit). Move cursor to the next line.
-        """
-        new_pos = self.document.cursor_down_position
-        if new_pos is not None:
-            self.cursor_position = new_pos
+        """ (for multiline edit). Move cursor to the next line.  """
+        self.cursor_position += self.document.get_cursor_down_position()
 
     @_to_mode(LineMode.NORMAL, LineMode.COMPLETE)
     def auto_up(self):
@@ -498,28 +490,6 @@ class Line(object):
         to_delete = len(self.document.current_line_before_cursor)
         return self.delete_character_before_cursor(to_delete)
 
-    @_to_mode(LineMode.NORMAL)
-    def delete_current_line(self):
-        """
-        Delete current line. Return deleted text.
-        """
-        document = self.document
-
-        # Remember deleted text.
-        deleted = document.current_line
-
-        # Cut line.
-        lines = document.lines
-        pos = document.cursor_position_row
-        self.text = '\n'.join(lines[:pos] + lines[pos+1:])
-
-        # Move cursor.
-        before_cursor = document.current_line_before_cursor
-        self.cursor_position -= len(before_cursor)
-        self.cursor_to_start_of_line(after_whitespace=True)
-
-        return deleted
-
     @_to_mode(LineMode.NORMAL)
     def join_next_line(self):
         """
@@ -542,30 +512,6 @@ class Line(object):
 
             self.text = self.text[:pos-2] + b + a + self.text[pos:]
 
-    @_to_mode(LineMode.NORMAL)
-    def go_to_matching_bracket(self):
-        """ Go to matching [, (, { or < bracket. """
-        stack = 1
-
-        for A, B in '()', '[]', '{}', '<>':
-            if self.document.current_char == A:
-                for i, c in enumerate(self.document.text_after_cursor[1:]):
-                    if c == A: stack += 1
-                    elif c == B: stack -= 1
-
-                    if stack == 0:
-                        self.cursor_position += (i + 1)
-                        break
-
-            elif self.document.current_char == B:
-                for i, c in enumerate(reversed(self.document.text_before_cursor)):
-                    if c == B: stack += 1
-                    elif c == A: stack -= 1
-
-                    if stack == 0:
-                        self.cursor_position -= (i + 1)
-                        break
-
     @_to_mode(LineMode.NORMAL)
     def go_to_substring(self, sub, in_current_line=False, backwards=False):
         """
@@ -579,18 +525,6 @@ class Line(object):
         if index:
             self.cursor_position += index
 
-    @_to_mode(LineMode.NORMAL)
-    def go_to_column(self, column):
-        """
-        Go to this column on the current line. (Go to the end column > length
-        of the line.)
-        """
-        line_length = len(self.document.current_line)
-        current_column = self.document.cursor_position_col
-        column = max(0, min(line_length, column))
-
-        self.cursor_position += column - current_column
-
     def create_code_obj(self):
         """
         Create `Code` instance from the current input.
@@ -667,8 +601,8 @@ class Line(object):
             self.complete_state = CompletionState(
                         original_document=self.document,
                         current_completions=current_completions)
-            self.insert_text(self.complete_state.current_completion_text)
             self.mode = LineMode.COMPLETE
+            self._go_to_completion(0)
 
         else:
             self.mode = LineMode.NORMAL
@@ -680,14 +614,11 @@ class Line(object):
         """
         assert self.mode == LineMode.COMPLETE
 
-        # Undo previous completion
-        count = len(self.complete_state.current_completion_text)
-        if count:
-            self.delete_character_before_cursor(count=len(self.complete_state.current_completion_text))
-
         # Set new completion
         self.complete_state.complete_index = index
-        self.insert_text(self.complete_state.current_completion_text)
+
+        # Set text/cursor position
+        self.text, self.cursor_position = self.complete_state.get_new_text_and_position()
 
         self.mode = LineMode.COMPLETE
 
@@ -802,35 +733,43 @@ class Line(object):
         self._clipboard = clipboard_data
 
     @_to_mode(LineMode.NORMAL)
-    def paste_from_clipboard(self, before=False):
+    def paste_from_clipboard(self, before=False, count=1):
         """
         Insert the data from the clipboard.
         """
         if self._clipboard and self._clipboard.text:
             if self._clipboard.type == ClipboardDataType.CHARACTERS:
                 if before:
-                    self.insert_text(self._clipboard.text)
+                    self.insert_text(self._clipboard.text * count)
                 else:
                     self.cursor_right()
-                    self.insert_text(self._clipboard.text)
+                    self.insert_text(self._clipboard.text * count)
                     self.cursor_left()
 
             elif self._clipboard.type == ClipboardDataType.LINES:
                 if before:
                     self.cursor_to_start_of_line()
-                    self.insert_text(self._clipboard.text + '\n', move_cursor=False)
+                    self.insert_text((self._clipboard.text + '\n') * count, move_cursor=False)
                 else:
                     self.cursor_to_end_of_line()
-                    self.insert_text('\n')
-                    self.insert_text(self._clipboard.text, move_cursor=False)
+                    self.insert_text(('\n' + self._clipboard.text) * count, move_cursor=False)
+                    self.cursor_down()
+
+                self.cursor_to_start_of_line(after_whitespace=True)
 
     @_to_mode(LineMode.NORMAL)
     def undo(self):
-        if self._undo_stack:
+        # Pop from the undo-stack until we find a text that if different from
+        # the current text. (The current logic of `save_to_undo_stack` will
+        # make sure that the top of the undo stack is usually the same as the
+        # current text, so in that case we have to pop twice.)
+        while self._undo_stack:
             text, pos = self._undo_stack.pop()
 
-            self.text = text
-            self.cursor_position = pos
+            if text != self.text:
+                self.text = text
+                self.cursor_position = pos
+                return
 
     @_to_mode(LineMode.NORMAL)
     def abort(self):
